/*    BURNIT ALL
 *    This simple program allows an ATMEL 2051, a PIC 16F628, or a GAL22V10 part to be burned 
 *    using an interface similar to the MINMON application used by MIT's 
 *    Microprocessor Project Laboratory (6.115).
 *
 *    Original 2051 work by Steve Whittaker with PIC and 22V10 additions by Zachary Clifford
 *
 *    Steve Whittaker, April-May 2007
 *    Zachary Clifford Jan-Aug 2008  
 */

#include <stdio.h>		/* used for string manipulation procedures */
#include <avr/io.h>		/* direct access to the AVRs IO ports/SFRs */
#include <avr/pgmspace.h>	/* for accessing strings in program memory */

#include <ctype.h>		/* to convert characters from upper to lowercase */

#include "avrutils.h"
#include "pic.h"		/*Constants for PIC programming */
#include "gal.h"
#include "2051.h"
#include "version.h"

//0 is ATMEL mode, 1 is PIC mode, 2 is 22V10 mode
static uint8_t currentMode = 0;

/*
 * Set up the UART using the cpu clock defined above.
 * 9600 Baud / 8-bit / 1-stop bit
 */
static void ioinit(void)
{
	//baud rate calculation
	UBRR0H = 0UL;
	UBRR0L = ((F_CPU / (16 * 9600UL)) - 1);	//9600 Bd (calculated using F_CPU,
	// defined above)

	UCSR0B = _BV(TXEN0) | _BV(RXEN0);	// tx/rx enable
	UCSR0C = _BV(UCSZ00) | _BV(UCSZ01);	// 8 bit bit
	UCSR0C &= ~(_BV(USBS0));	// 1 stop bit
}

/*    
 *    Reads in a Intel HEX file over the serial port and 
 *    stores the data in memory
 *    Also understands if PIC is being used.
 *
 *    Some hex processing hints taken from AVRFreaks.net forums.
 */
static uint8_t load_ihex()
{
	printstr_p(PSTR("\n>"));	//the programming prompt
	uint8_t data_pairs, address_hi, address_lo, temp_byte, i, checksum,
	    type;

	while (1) {

		while (getchr() != ':')	// go forward until we get a colon
		{
			;
		}

		/*get the number of ascii character pairs on this line */
		data_pairs = ascii_to_bin(getchr()) << 4;
		data_pairs |= ascii_to_bin(getchr());

		/*address to write to */
		address_hi = ascii_to_bin(getchr()) << 4;
		address_hi |= ascii_to_bin(getchr());

		address_lo = ascii_to_bin(getchr()) << 4;
		address_lo |= ascii_to_bin(getchr());

		/* get the data type */
		type = ascii_to_bin(getchr()) << 4;
		type |= ascii_to_bin(getchr());

		checksum = (type + address_lo + address_hi + data_pairs) & 0xFF;

		if (type != 0 && type != 1) {
			//Something is wrong
			//Borrowed error handling from JimPic by Jim Paris
			if (type == 0x04 || type == 0x02) {
				/* Ignore this one silently; used for 32-bit addressing */
			} else {
				printstr_p(PSTR
					   ("\nWarning: ignoring HEX record "));
				char s[20];
				sprintf(s, "type %02x\n", type);
				printstr(s);
			}
			continue;
		}

		for (i = 0; i < data_pairs; i++) {
			temp_byte = ascii_to_bin(getchr()) << 4;
			temp_byte |= ascii_to_bin(getchr());
			checksum += temp_byte;
			if ((address_lo + 256 * address_hi + i <= 0x400F) &&
			    (address_lo + 256 * address_hi + i >= 0x4000)) {
				//It's PIC configuration bytes
				picConfigMemory[address_lo + i] = temp_byte;
			} else if ((address_lo + 256 * address_hi + i >= 0x4200)
				   && (address_lo + 256 * address_hi + i <=
				       0x4280)) {
				//It's PIC Data.  That usually lives in 0x2100 to 0x2180
				//Doubled to 0x4200 to 0x4280
				picDataMemory[address_lo + i] = temp_byte;
			} else if (address_lo + 256 * address_hi + i >=
				   CHIPMEMSIZE) {
				//It is bigger than normal data but not any of the 
				//other special cases
				printstr_p(PSTR
					   ("\nERROR: This hex file has data in high memory"
					    " that does not correspond\n"));
				printstr_p(PSTR
					   ("to Data or Config memory.  It might also be"
					    " too big for this chip."));
				return -1;
			} else {
				//It's plain old data.
				if ((currentMode == 1) &&
				    ((PICtype == PIC16F627)
				     || (PICtype == PIC16F627A))
				    && ((uint16_t)
					(address_lo + 256 * address_hi + i) >
					1024)) {
					//It is a small PIC and we're writing outside
					//its implemented memory
					printstr_p(PSTR
						   ("\nERROR: This hex file has more data"
						    " than the PIC16F627 can hold."));
					printstr_p(PSTR
						   ("\nCheck your compiler settings or get a 16F628."));
					return -1;
				}
				chipMemory[address_lo + 256 * address_hi + i] =
				    temp_byte;
			}
		}

		/*and the checksum */
		temp_byte = ascii_to_bin(getchr()) << 4;
		temp_byte |= ascii_to_bin(getchr());

		if (((checksum + temp_byte) & 0xFF) != 0) {
			printstr_p(PSTR
				   ("\nERROR: Checksum invalid in hex file."));
		}
		//if it's the end of record and no data, terminate
		if (type == 1 && data_pairs == 0) {
			getchr();
			break;
		}

		printstr_p(PSTR("."));	/*display progress */
	}

	return 0;

}

/*
 * initPins - Called at the beginning of BurnIt.  Because we might have a GAL, pull-ups should be active on
 * all outputs.  Shuts off all output pins
 * and ensures that high voltage pins are inactive
 */
void initPins(void)
{

	//Ensure pullups are on
	MCUCR &= ~(_BV(PUD));

	//Set all pins to inputs except for the high voltage buffers at PB1 and PD6
	//Keep those low to keep from applying the high voltage.  Pull others high with weak pullups.
	//Setting the pins in this way should not interfere with the serial port because of internal override signals
	DDRA = 0x00;
	RSTDDR = _BV(RST12);
	DDRC = 0x00;
	EDITDDR = _BV(EDIT);

	RSTPORT = ~(_BV(RST12));
	PORTA = 0xFF;
	PORTC = 0xFF;
	EDITPORT = ~(_BV(EDIT));

}

    /* Check for which mode we're in.
     * The DP3T switch applies ground to different pins.
     * BURNIT senses the GND pins of the DP3T switch.  ATMEGA pins will have
     * weak pullups to VCC so they are 1 if in that position and 0 otherwise.
     * If switch is "right"
     * we're in PIC mode and the PIC VSS will be grounded.  PIC VCC will be
     * applied as well.  If switch is "center" VCC will be applied for 2051 and
     * other part of the switch will be grounded.
     * If switch is "left", no power is applied. The GND side pulls down on a pin
     * for sensing purposes.
     */

int getInsertedChip(void)
{
	if ((GALSENSEPIN & _BV(GALSENSE)) == 0) {
		//GALSENSE was pulled low, so it's a GAL
		return 2;
	}
	if ((PICSENSEPIN & _BV(PICSENSE)) == 0) {
		//It's a PIC
		return 1;
	}
	if ((SENSE2051PIN & _BV(SENSE2051)) == 0) {
		//It's a 2051
		return 0;
	}
	//We don't know for sure.  However, BurnIt and PICBurnIt do not have a SENSE2051PIN
	//So that those boards work, we will defaut to 2051.
	return 0;
}

/*
 * The BurnIT main program loop
 */

int main(void)
{

//      short location;

	//For all chips, ensure that drivers are off and that high voltage is
	//not active.
	initPins();
	ioinit();		//set up the serial port

	getchr();		//wait for first keypress.
	printstr_p(PSTR("\nWelcome to BURNIT v" VERSION "!\nPress 'H' for help."));

	//Now detect the chip and properly initialize for that chip
	currentMode = getInsertedChip();

	if (currentMode == 1) {
		printstr_p(PSTR("\nPIC mode enabled"));

		//Set up port pins for PIC mode
		DDRB = PORTB_PIC_DDR;
		DDRA = PORTA_PIC_DDR;

		pic_get_revision();

		if (PICtype == 0) {
			printstr_p(PSTR("\nMaybe the chip isn't inserted or "
					"the switch is in the wrong position."));
			printstr_p(PSTR("\nPlease correct and reset BurnIt\n"));
			while (1) ;
		}
	} else if (currentMode == 0) {
		printstr_p(PSTR("\nAtmel 2051 mode enabled"));

		//standard setup, data is output, everything zeroed.
		DATA_DDR = 0xFF;	//Data is output
		CONTROL_DDR = 0xFF;	//Control Port is all output.
		FEEDBACK_DDR = 0x00;	//Feedback is inputs

		CONTROL_PORT = 0x00;
		DATA_PORT = 0x00;

		atmel_pulse_xtal();
	} else if (currentMode == 2) {
		printstr_p(PSTR("\nGAL22V10 mode enabled"));
		//Leave the pins alone for now
		//Because we do not know how the GAL is presently programmed, we do not know what are inputs and outputs
		//We cannot activate the drivers until EDIT is asserted
	} else {
		printstr_p(PSTR
			   ("\nUnrecognized chip, please ensure it is inserted and try again"));
		while (1) ;
	}

	/*initialize memory to 0x00 if necessary for chip type */
	int ix;
	if (currentMode == 0) {
		for (ix = 0; ix < CHIPMEMSIZE; ix++) {
			chipMemory[ix] = 0x00;
		}
	} else if (currentMode == 1) {
		//We're in PIC mode.  Initialize to 0xFF3F repeating
		//That way we know which words need programming.
		for (ix = 0; ix < CHIPMEMSIZE / 2; ix++) {
			chipMemory[2 * ix] = 0xFF;
			chipMemory[2 * ix + 1] = 0x3F;
		}
		for (ix = 0; ix < 128; ix++) {
			picDataMemory[ix] = 0xFF;
		}
		for (ix = 0; ix < 8; ix++) {
			picConfigMemory[2 * ix] = 0xFF;
			picConfigMemory[2 * ix + 1] = 0x3F;
		}
	} else if (currentMode == 2) {
		//We're in GAL mode, so no init is necessary
		//The JEDEC file determines whether the default is zeros or ones
	}

	/*Main Loop */
	while (1) {
		printstr_p(PSTR("\nBURNIT>"));	/*display prompt */
		char rxbyte = toupper(getchr());	/*get character from serial, convert to upper */
		putchr(rxbyte);
		if (currentMode == 1) {
			switch (rxbyte) {
			case 'C':
				pic_bulk_erase();
				break;
			case 'D':
				load_ihex();
				break;

			case 'A':
				load_ihex();	//Fallthrough intentional
			case 'P':

				pic_bulk_erase();

				if (pic_write_program() < 0)
					break;

				if (pic_write_data() < 0)
					break;

				pic_write_configuration();

				break;
			case 'V':
				pic_view();
				break;
			case 'U':
				pic_total_erase();
				break;
			case 'H':
				printstr_p(PSTR("\nBURNIT Help"));
				printstr_p(PSTR("\nP    Program a PIC."));
				printstr_p(PSTR("\nC    Clear a PIC."));
				printstr_p(PSTR
					   ("\nD    Download an Intel Hex file to BURNIT."));
				printstr_p(PSTR
					   ("\nA    Auto Download and Program PIC."));
				printstr_p(PSTR
					   ("\nH    Print this help document."));
				printstr_p(PSTR
					   ("\nV    View the code on the currently inserted PIC."));
				printstr_p(PSTR
					   ("\nU    Unlock PIC if code protection enabled."));
				break;
			case '\r':	//FALLTHROUGH intentional
			case '\n':
				break;
			default:
				printstr_p(PSTR("\ninvalid command"));
			}

		} else if (currentMode == 2) {
			//Show GAL menu
			switch (rxbyte) {
			case 'C':
				//ReadPES();
				ParsePES(0);
				printstr_p(PSTR("\nErasing GAL..."));
				EraseGAL();
				printstr_p(PSTR("\nGAL Erased"));
				break;
			case 'D':
				if (readJEDEC()) {
					printstr_p(PSTR
						   ("\nJEDEC Download Complete"));
				} else {
					printstr_p(PSTR
						   ("\nERROR in downloading JEDEC"));
				}
				break;

			case 'A':
				if (readJEDEC()) {
					printstr_p(PSTR
						   ("\nJEDEC Download Complete"));
				} else {
					printstr_p(PSTR
						   ("\nERROR in downloading JEDEC...Autoprogramming will not continue"));
					break;
				}	//Fallthrough intentional
			case 'P':
				//ReadPES(); //Get the programmer's signature
				ParsePES(0);	//Parse it
				printstr_p(PSTR("\nErasing GAL..."));
				EraseGAL();
				printstr_p(PSTR("\nGAL Erased"));
				printstr_p(PSTR("\nProgramming GAL..."));
				//WriteGAL(1);
				//EraseGAL();
				WriteGAL(0);
				printstr_p(PSTR
					   ("\nGAL Programmed...Verifying..."));
				//location = ReadGAL(1);
				if (ReadGAL(1)) {
					printstr_p(PSTR
						   ("\nVerification Error"));
					//char s[10];
					//sprintf(s,"%hd",location);
					//printstr(s);
				} else {
					printstr_p(PSTR("\nVerification OK."));
				}
				break;
			case 'V':
				//ReadPES(); //Get the programmer's signature
				ParsePES(0);	//Parse it
				ReadGAL(0);
				PrintFuses();
				break;
			case 'T':
				ReadPES();
				ParsePES(1);
				break;
			case 'H':
				printstr_p(PSTR("\nBURNIT Help"));
				printstr_p(PSTR("\nP    Program a GAL."));
				printstr_p(PSTR("\nC    Clear a GAL."));
				printstr_p(PSTR
					   ("\nD    Download a JEDEC file to BURNIT."));
				printstr_p(PSTR
					   ("\nA    Auto Download and Program a GAL."));
				printstr_p(PSTR
					   ("\nV    View Fusemap of inserted GAL."));
				//printstr_p(PSTR("\nT    Test view the PES."));
				printstr_p(PSTR
					   ("\nH    Print this help document."));
				break;
			case '\r':	//FALLTHROUGH intentional
			case '\n':
				break;
			default:
				printstr_p(PSTR("\ninvalid command"));
			}

		} else {

			switch (rxbyte) {
			case 'P':
				atmel_program();
				break;
			case 'V':
				atmel_view();
				break;
			case 'C':
				atmel_clear_chip();
				break;
			case 'A':
				load_ihex();
				atmel_program();
				break;
			case 'D':
				load_ihex();
				break;
				//case 'S':
				//    atmel_signature();
				//    break;
			case 'H':
				printstr_p(PSTR("\nBURNIT Help"));
				printstr_p(PSTR
					   ("\nP    Program an ATMEL 2051/4051."));
				printstr_p(PSTR
					   ("\nC    Clear an ATMEL 2051/4051."));
				printstr_p(PSTR
					   ("\nD    Download an Intel Hex file to BURNIT."));
				printstr_p(PSTR
					   ("\nA    Auto Download and Program 2051/4051."));
				printstr_p(PSTR
					   ("\nH    Print this help document."));
				printstr_p(PSTR
					   ("\nV    View the code on the currently inserted 2051/4051."));
				//printstr_p(PSTR("\nS    Read the signature bits of the chip."));                
				break;
			case '\r':	//FALLTHROUGH intentional
			case '\n':
				break;
			default:
				printstr_p(PSTR("\ninvalid command"));
			}
		}
	}

}
