
#include <stdio.h>		/* used for string manipulation procedures */
#include <avr/io.h>		/* direct access to the AVRs IO ports/SFRs */
#include <avr/pgmspace.h>	/* for accessing strings in program memory */

#include <ctype.h>		/* to convert characters from upper to lowercase */

#include "avrutils.h"
#include "pic.h"		/*Constants for PIC programming */

//sets type of PIC
uint8_t PICtype = 0;

//Data memory for the PIC to be burned
uint8_t picDataMemory[128];

//Configuration memory for the PIC to be burned
uint8_t picConfigMemory[16];

/*
 * The following are low level routines to help portability to other platforms
 */
static void pic_set_clock(int setClock)
{
	if (setClock) {
		CLOCK_PORT |= _BV(CLOCK);
	} else {
		CLOCK_PORT &= ~(_BV(CLOCK));
	}
}

static void pic_set_pgm(int setPGM)
{
	if (setPGM) {
		PGM_PORT |= _BV(PGM);
	} else {
		PGM_PORT &= ~(_BV(PGM));
	}
}

static void pic_set_vpp(int setVPP)
{
	if (setVPP) {
		VPP_PORT |= _BV(VPP);
	} else {
		VPP_PORT &= ~(_BV(VPP));
	}
}

/*
 * pic_enter_programming
 * Enter PIC programming mode
 *
 */
static void pic_enter_programming(void)
{

	//Ensure everything is low
	pic_set_clock(0);
	pic_set_pgm(0);
	pic_set_vpp(0);

	//Wait delay between power on and PGM rise
	_delay_us(5);

	//Now raise lines
	pic_set_pgm(1);
	_delay_us(5);

	pic_set_vpp(1);
	_delay_us(5);

	//Now we are ready for command
}

/*
 * pic_exit_programming
 * Exit PIC programming mode
 */
static void pic_exit_programming(void)
{
	pic_set_vpp(0);
	pic_set_pgm(0);
}

/*
 * pic_pulse_clock
 * Pulses the clock to the PIC
 */
static void pic_pulse_clock(void)
{
	//Clock needs setup and hold of 100ns each.  Both should be met
	//without explicit delays.

	_delay_us(2);
	pic_set_clock(0);

	_delay_us(2);

	pic_set_clock(1);

	_delay_us(2);
}

/*
 * pic_send_command
 * Sends the specified 6 bit command to the PIC
 * Assumes already in programming mode
 */
static void pic_send_command(uint8_t comm)
{

	//Make DATA an output
	DATA_DDR |= _BV(DATADDR);

	//Raise the clock line
	pic_set_clock(1);

	DATA_PORT = comm & 0x1 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x2 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x4 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x8 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x10 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x20 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));

	//Don't pulse last clock because it needs to remain low for a while.
	pic_set_clock(0);

	//Shut off the data port before the PIC makes an attept to drive it.
	//Make an input.
	DATA_DDR &= ~(_BV(DATADDR));

	//Delay after command sent
	_delay_us(1);
}

/*
 * pic_receive_data
 * Receives 16 bits of data and returns 14 bits actually sent.
 */
static uint16_t pic_receive_data(void)
{

	uint16_t data = 0;
	//Make DATA an input
	DATA_DDR &= ~(_BV(DATADDR));

	//Raise the clock line
	pic_set_clock(1);

	//Pulse clock to receive leading zero
	pic_pulse_clock();

	//Now get the data
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x1 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x2 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x4 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x8 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x10 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x20 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x40 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x80 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x100 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x200 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x400 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x800 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x1000 : 0);
	pic_pulse_clock();
	data |= ((DATA_PIN & _BV(DATAIN)) ? 0x2000 : 0);
	pic_pulse_clock();

	//Now have the data.  Just lower clock one more time to clear lagging zero
	pic_set_clock(0);

	_delay_us(1);

	return data;
}

/*
 * pic_send_data
 * Sends the specified 14 bit command (will be 0 padded to 16) to the PIC
 * Assumes already in programming mode and command already sent
 */
static void pic_send_data(uint16_t comm)
{
	//Make DATA an output
	DATA_DDR |= _BV(DATADDR);

	//Raise the clock line
	pic_set_clock(1);

	//Send 0

	DATA_PORT = DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();

	DATA_PORT = comm & 0x1 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x2 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x4 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x8 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x10 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x20 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x40 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x80 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x100 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x200 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x400 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x800 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x1000 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();
	DATA_PORT = comm & 0x2000 ?
	    DATA_PORT | _BV(DATA) : DATA_PORT & ~(_BV(DATA));
	pic_pulse_clock();

	//Send trailing zero
	DATA_PORT = DATA_PORT & ~(_BV(DATA));
	//Don't pulse last clock because it needs to remain low for a while.
	pic_set_clock(0);

	//Shut off the data port
	//Make an input.
	DATA_DDR &= ~(_BV(DATADDR));

	//Delay after data sent
	_delay_us(1);
}

/*
 * pic_bulk_erase
 * Erase the PIC's memory
 */
void pic_bulk_erase(void)
{
	pic_enter_programming();
	printstr_p(PSTR("\nErasing PIC program memory..."));

	//First load data for program memory with data set to all ones.
	pic_send_command(LOAD_PROGRAM_COMMAND);

	pic_send_data(0x3FFF);

	pic_send_command(BULK_ERASE_PROGRAM);

	if ((PICtype == PIC16F628) || (PICtype == PIC16F627)) {
		pic_send_command(BEGIN_PROGRAM_ONLY);
	}

	_delay_ms(6);

	printstr_p(PSTR("\nErasing PIC data memory..."));

	pic_send_command(LOAD_PROGRAM_COMMAND);

	pic_send_data(0x3FFF);

	pic_send_command(BULK_ERASE_DATA);

	if ((PICtype == PIC16F628) || (PICtype == PIC16F627)) {
		pic_send_command(BEGIN_PROGRAM_ONLY);
	}

	_delay_ms(6);
	pic_exit_programming();
}

/*
 * pic_write_program
 * Program PIC memory
 */
int pic_write_program(void)
{
	uint16_t i = 0;
	uint16_t tempdata;
	uint16_t sendingdata;

	pic_enter_programming();
	printstr_p(PSTR("\nProgramming PIC Program memory..."));

	for (i = 0; i < CHIPMEMSIZE / 2; i++) {
		if (i % 32 == 0) {
			printstr_p(PSTR("."));
		}

		sendingdata = (((chipMemory[2 * i + 1] << 8) & 0x3F00) |
			       chipMemory[2 * i]) & 0x3FFF;

		//Don't bother sending data if it's blank.
		if ((sendingdata == 0x3FFF)) {
			pic_send_command(INCREMENT_ADDRESS);
			continue;
		}
		//Load data for program memory
		pic_send_command(LOAD_PROGRAM_COMMAND);
		//Send the data
		pic_send_data((((chipMemory[2 * i + 1] << 8) & 0x3F00) |
			       chipMemory[2 * i]) & 0x3FFF);
		//Begin programming only cycle
		if ((PICtype == PIC16F628) || (PICtype == PIC16F627)) {
			pic_send_command(BEGIN_PROGRAM_ONLY);
			_delay_ms(8);
		} else {
			pic_send_command(A_PROGRAM_ONLY);
			_delay_ms(4);
		}

		//Now verify that it was written
		//Read from Program
		pic_send_command(READ_PROGRAM_COMMAND);
		tempdata = pic_receive_data();
		if ((tempdata & 0x3FFF) !=
		    ((((chipMemory[2 * i + 1] << 8) & 0x3F00) |
		      chipMemory[2 * i]) & 0x3FFF)) {
			printstr_p(PSTR
				   ("\nVerification error during programming."));
			printstr_p(PSTR("\nCode protection may be enabled."));
			pic_exit_programming();
			return -1;
		}
		//Increment address
		pic_send_command(INCREMENT_ADDRESS);
	}
	pic_exit_programming();
	return 0;
}

/*
 * pic_write_data
 * Program PIC memory
 */
int pic_write_data(void)
{
	uint8_t i = 0;
	uint16_t tempdata;
	pic_enter_programming();
	printstr_p(PSTR("\nProgramming PIC Data memory..."));

	for (i = 0; i < 128; i++) {
		if (picDataMemory[i] == 0xFF) {
			pic_send_command(INCREMENT_ADDRESS);
			continue;
		}
		//Load data for program memory
		pic_send_command(LOAD_DATA_COMMAND);
		//Send the data
		pic_send_data(picDataMemory[i] & 0x00FF);
		//Begin programming only cycle
		if ((PICtype == PIC16F628) || (PICtype == PIC16F627)) {
			pic_send_command(BEGIN_PROGRAM_ONLY);
			_delay_ms(8);
		} else {
			pic_send_command(A_PROGRAM_ONLY);
			_delay_ms(4);
		}

		//Now verify that it was written
		//Read from Data
		pic_send_command(READ_DATA_COMMAND);
		tempdata = pic_receive_data();
		if ((tempdata & 0x00FF) != (picDataMemory[i] & 0x00FF)) {
			printstr_p(PSTR
				   ("\nVerification error during Data programming."));
			pic_exit_programming();
			return -1;
		}
		//Increment address
		pic_send_command(INCREMENT_ADDRESS);
	}
	pic_exit_programming();
	return 0;
}

/*
 * pic_write_configuration
 * Program PIC config memory
 */
int pic_write_configuration(void)
{
	uint8_t i = 0;
	uint16_t tempdata;
	uint16_t sendingdata;
	printstr_p(PSTR("\nProgramming PIC Configuration Word..."));
	pic_enter_programming();
	pic_send_command(LOAD_CONFIG_COMMAND);
	pic_send_data(0x3FFF);

	for (i = 0; i < 8; i++) {
		if (i > 3 && i < 7) {
			pic_send_command(INCREMENT_ADDRESS);
			continue;
		}
		sendingdata = (((picConfigMemory[2 * i + 1] << 8) & 0x3F00) |
			       picConfigMemory[2 * i]) & 0x3FFF;

		//Don't bother sending data if it's blank.
		if ((sendingdata == 0x3FFF)) {
			pic_send_command(INCREMENT_ADDRESS);
			continue;
		}
		//Load data for program memory
		pic_send_command(LOAD_PROGRAM_COMMAND);
		//Send the data
		pic_send_data((((picConfigMemory[2 * i + 1] << 8) & 0x3F00) |
			       picConfigMemory[2 * i]) & 0x3FFF);
		//Begin programming only cycle
		if ((PICtype == PIC16F628) || (PICtype == PIC16F627)) {
			//Using Erase/Program because the bulk erase apparently doesn't
			//clear this memory space.
			pic_send_command(BEGIN_ERASE_PROGRAM);
			_delay_ms(13);
		} else {
			pic_send_command(A_PROGRAM_ONLY);
			_delay_ms(4);
		}

		//Now verify that it was written
		//Read from Program
		pic_send_command(READ_PROGRAM_COMMAND);
		tempdata = pic_receive_data();
		if ((tempdata & 0x3FFF) !=
		    ((((picConfigMemory[2 * i + 1] << 8) & 0x3F00) |
		      picConfigMemory[2 * i]) & 0x3FFF)) {
			printstr_p(PSTR
				   ("\nVerification error during Configuration"
				    " programming."));
			printstr_p(PSTR("\n Sent: "));
			printhex(picConfigMemory[2 * i + 1]);
			printhex(picConfigMemory[2 * i]);
			printstr_p(PSTR("\n Read: "));
			printhex((tempdata >> 8) & 0x00FF);
			printhex(tempdata & 0x00FF);

			printstr_p(PSTR
				   ("\nNOTE: This programmer cannot disable Low "
				    "Voltage Programming\n"));
			printstr_p(PSTR
				   ("If that was the only problem, do not worry."));
			pic_exit_programming();
			return -1;
		}
		//Increment address
		pic_send_command(INCREMENT_ADDRESS);

	}
	pic_exit_programming();
	return 0;
}

/*
 * pic_get_revision
 * Gets the revision of this PIC and prints it for debugging right now.
 */
uint8_t pic_get_revision(void)
{
	uint16_t deviceid;
	pic_enter_programming();

	//Load configuration
	pic_send_command(0x00);
	pic_send_data(0x0000);

	pic_send_command(0x06);
	pic_send_command(0x06);
	pic_send_command(0x06);
	pic_send_command(0x06);
	pic_send_command(0x06);
	pic_send_command(0x06);

	//Should be advanced to 0x2006 by now
	//Get data
	pic_send_command(0x04);

	deviceid = pic_receive_data();

	//Mask off the lower 5 bits because they are just revision.
	deviceid = (deviceid >> 5) & 0xFF;

	switch (deviceid) {
	case PIC16F648A:
		printstr_p(PSTR("\nPIC16F648A detected"));
		printstr_p(PSTR("\nNote: This programmer can only program "
				"the first 2K of memory and not the whole"));
		printstr_p(PSTR("\n4K of memory on this device."));
		printstr_p(PSTR("\nIt can also only program the first 128 bytes"
				" of Data and not all 256 bytes."));
		PICtype = PIC16F648A;
		return PIC16F648A;
		break;
	case PIC16F627A:
		printstr_p(PSTR("\nPIC16F627A detected"));
		PICtype = PIC16F627A;
		return PIC16F627A;
		break;
	case 0x83:
		printstr_p(PSTR("\nPIC16F628A detected"));
		PICtype = PIC16F628A;
		return PIC16F628A;
		break;
	case PIC16F627:
		printstr_p(PSTR("\nPIC16F627 detected"));
		PICtype = PIC16F627;
		return PIC16F627;
		break;
	case 0x3E:
		printstr_p(PSTR("\nPIC16F628 detected"));
		PICtype = PIC16F628;
		return PIC16F628;
		break;
	default:
		printstr_p(PSTR("\nUnknown PIC, Config word was:"));
		printhex((deviceid >> 8) & 0xFF);
		printhex(deviceid & 0xFF);
		return deviceid;
		break;
	}

	pic_exit_programming();
}

/*
 * pic_view
 * view the code on the PIC
 */
void pic_view(void)
{
	pic_enter_programming();

	pic_send_command(LOAD_PROGRAM_COMMAND);
	pic_send_data(0x0000);

	int16_t ix;
	for (ix = 0; ix < CHIPMEMSIZE / 2; ix++) {
		if ((ix % 16) == 0) {	/*line numbers and newlines */
			printstr_p(PSTR("\n"));
			char line_num[10];
			sprintf(line_num, "%04X :\t", (2 * ix));
			printstr(line_num);

		}
		pic_send_command(READ_PROGRAM_COMMAND);
		uint16_t curr_mem = pic_receive_data();
		printhex(curr_mem & 0xFF);
		printhex((curr_mem >> 8) & 0xFF);
		pic_send_command(INCREMENT_ADDRESS);
	}

	pic_exit_programming();
	pic_enter_programming();

	printstr_p(PSTR("\n DATA :"));
	for (ix = 0; ix < 128; ix++) {
		if ((ix % 32) == 0) {	/*line numbers and newlines */
			printstr_p(PSTR("\n"));
			char line_num[10];
			sprintf(line_num, "%04X :\t", (ix));
			printstr(line_num);
		}
		pic_send_command(READ_DATA_COMMAND);

		uint16_t curr_mem = pic_receive_data();
		printhex(curr_mem & 0x00FF);

		pic_send_command(INCREMENT_ADDRESS);
	}

	pic_exit_programming();
	pic_enter_programming();

	pic_send_command(LOAD_CONFIG_COMMAND);
	pic_send_data(0x3FFF);

	printstr_p(PSTR("\n USER ID : "));

	for (ix = 0; ix < 4; ix++) {

		pic_send_command(READ_PROGRAM_COMMAND);
		uint16_t curr_mem = pic_receive_data();
		printhex(curr_mem & 0xFF);
		printhex((curr_mem >> 8) & 0xFF);
		pic_send_command(INCREMENT_ADDRESS);

	}

	pic_send_command(INCREMENT_ADDRESS);
	pic_send_command(INCREMENT_ADDRESS);
	pic_send_command(INCREMENT_ADDRESS);

	printstr_p(PSTR("\n CONFIG WORD : "));

	pic_send_command(READ_PROGRAM_COMMAND);
	uint16_t curr_mem = pic_receive_data();
	printhex(curr_mem & 0xFF);
	printhex((curr_mem >> 8) & 0xFF);

	pic_exit_programming();
}

/*
 * pic_total_erase
 * Perform total PIC bulk erase.  Necessary to clear data protection bits.
 */
void pic_total_erase(void)
{

	printstr_p(PSTR("\nPerforming total bulk erase to "
			"clear code protection..."));
	pic_enter_programming();

	if ((PICtype == PIC16F628) || (PICtype == PIC16F627)) {
		pic_send_command(LOAD_CONFIG_COMMAND);
		pic_send_data(0x3FFF);

		pic_send_command(INCREMENT_ADDRESS);
		pic_send_command(INCREMENT_ADDRESS);
		pic_send_command(INCREMENT_ADDRESS);
		pic_send_command(INCREMENT_ADDRESS);
		pic_send_command(INCREMENT_ADDRESS);
		pic_send_command(INCREMENT_ADDRESS);
		pic_send_command(INCREMENT_ADDRESS);

		pic_send_command(BULK_ERASE_SETUP_ONE);
		pic_send_command(BULK_ERASE_SETUP_TWO);

		pic_send_command(BEGIN_ERASE_PROGRAM);

		_delay_ms(13);

		pic_send_command(BULK_ERASE_SETUP_ONE);
		pic_send_command(BULK_ERASE_SETUP_TWO);
	} else {

		pic_send_command(LOAD_CONFIG_COMMAND);

		pic_send_data(0x3FFF);

		pic_send_command(BULK_ERASE_PROGRAM);

		_delay_ms(6);
	}

	pic_exit_programming();
}
